#  SPDX-License-Identifier: Apache-2.0
"""
Python Package for connecting to myenergi API.

"""
import logging
import sys
from typing import Text

import httpx
from pycognito import Cognito

from .exceptions import MyenergiException
from .exceptions import TimeoutException
from .exceptions import WrongCredentials

_LOGGER = logging.getLogger(__name__)
_USER_POOL_ID = "eu-west-2_E57cCJB20"
_CLIENT_ID = "2fup0dhufn5vurmprjkj599041"


class Connection:
    """Connection to myenergi API."""

    def __init__(
        self,
        username: Text = None,
        password: Text = None,
        app_password: Text = None,
        app_email: Text = None,
        timeout: int = 20,
        asyncClient=httpx.AsyncClient(),
    ) -> None:
        """Initialize connection object."""
        self.timeout = timeout
        self.director_url = "https://director.myenergi.net"
        self.base_url = None
        self.asyncClient = asyncClient
        self.oauth_base_url = "https://myaccount.myenergi.com"
        self.username = username
        self.password = password
        self.app_password = app_password
        self.app_email = app_email
        self.auth = httpx.DigestAuth(self.username, self.password)
        self.headers = {"User-Agent": "Wget/1.14 (linux-gnu)"}
        if self.app_email and self.app_password:
            self.oauth = Cognito(_USER_POOL_ID, _CLIENT_ID, username=self.app_email)
            self.oauth.authenticate(password=self.app_password)
            self.oauth_headers = {"Authorization": f"Bearer {self.oauth.access_token}"}
        self.do_query_asn = True
        self.invitation_id = ""
        _LOGGER.debug("New connection created")

    def _checkMyenergiServerURL(self, responseHeader):
        if "X_MYENERGI-asn" in responseHeader:
            new_url = "https://" + responseHeader["X_MYENERGI-asn"]
            if new_url != self.base_url:
                _LOGGER.info(f"Updated myenergi active server to {new_url}")
            self.base_url = new_url
        else:
            _LOGGER.debug(
                "Myenergi ASN not found in Myenergi header, assume auth failure (bad username)"
            )
            raise WrongCredentials()

    async def discoverLocations(self):
        if self.app_email and self.app_password:
            locs = await self.get("/api/Location", oauth=True)
            # check if guest location - use the first location by default
            if locs["content"][0]["isGuestLocation"]:
                self.invitation_id = locs["content"][0]["invitationData"][
                    "invitationId"
                ]

    def checkAndUpdateToken(self):
        # check if we have oauth credentials
        if self.app_email and self.app_password:
            # check if we have to renew out token
            self.oauth.check_token()
            self.oauth_headers = {"Authorization": f"Bearer {self.oauth.access_token}"}

    async def send(self, method, url, json=None, oauth=False):
        # Use OAuth for myaccount.myenergi.com
        if oauth:
            # check if we have oauth credentials
            if self.app_email and self.app_password:
                theUrl = self.oauth_base_url + url
                # if we have an invitiation id, we need to add that to the query
                if self.invitation_id != "":
                    if "?" in theUrl:
                        theUrl = theUrl + "&invitationId=" + self.invitation_id
                    else:
                        theUrl = theUrl + "?invitationId=" + self.invitation_id
                try:
                    _LOGGER.debug(f"{method} {url} {theUrl}")
                    response = await self.asyncClient.request(
                        method,
                        theUrl,
                        json=json,
                        headers=self.oauth_headers,
                        timeout=self.timeout,
                    )
                except httpx.ReadTimeout:
                    raise TimeoutException()
                else:
                    _LOGGER.debug(f"{method} status {response.status_code}")
                    if response.status_code == 200:
                        return response.json()
                    elif response.status_code == 401:
                        raise WrongCredentials()
                    raise MyenergiException(response.status_code)
            else:
                _LOGGER.error("Trying to use OAuth without app credentials")

        # Use Digest Auth for director.myenergi.net and s18.myenergi.net
        else:
            # If base URL has not been set, make a request to director to fetch it
            if self.base_url is None or self.do_query_asn:
                _LOGGER.debug("Get Myenergi base url from director")
                try:
                    directorUrl = self.director_url + "/cgi-jstatus-E"
                    response = await self.asyncClient.get(
                        directorUrl,
                        auth=self.auth,
                        headers=self.headers,
                        timeout=self.timeout,
                    )
                except Exception:
                    _LOGGER.error("Myenergi server request problem")
                    _LOGGER.debug(sys.exc_info()[0])
                else:
                    self.do_query_asn = False
                    self._checkMyenergiServerURL(response.headers)
            theUrl = self.base_url + url
            try:
                _LOGGER.debug(f"{method} {url} {theUrl}")
                response = await self.asyncClient.request(
                    method,
                    theUrl,
                    auth=self.auth,
                    headers=self.headers,
                    timeout=self.timeout,
                    json=json,
                )
            except httpx.ReadTimeout:
                # Make sure to query for ASN next request, might be a server problem
                self.do_query_asn = True
                raise TimeoutException()
            else:
                _LOGGER.debug(f"GET status {response.status_code}")
                self._checkMyenergiServerURL(response.headers)
                if response.status_code == 200:
                    return response.json()
                elif response.status_code == 401:
                    raise WrongCredentials()
                # Make sure to query for ASN next request, might be a server problem
                self.do_query_asn = True
                raise MyenergiException(response.status_code)

    async def get(self, url, data=None, oauth=False):
        return await self.send("GET", url, data, oauth)

    async def post(self, url, data=None, oauth=False):
        return await self.send("POST", url, data, oauth)

    async def put(self, url, data=None, oauth=False):
        return await self.send("PUT", url, data, oauth)

    async def delete(self, url, data=None, oauth=False):
        return await self.send("DELETE", url, data, oauth)



from unittest.mock import MagicMock

import pytest

from pymyenergi.client import MyenergiClient
from pymyenergi.eddi import Eddi
from pymyenergi.harvi import Harvi
from pymyenergi.libbi import Libbi
from pymyenergi.zappi import Zappi

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class MockConnection:
    """Connection to myenergi API."""

    def __init__(self, app_password, app_email) -> None:
        """Initialize connection object."""
        self.app_password = app_password
        self.app_email = app_email

    async def discoverLocations(self):
        return

    def checkAndUpdateToken(self):
        return

    async def send(self, method, url, json=None, oauth=False):
        return MagicMock()

    async def get(self, url, data=None, oauth=False):
        return await self.send("GET", url, data, oauth)

    async def post(self, url, data=None, oauth=False):
        return await self.send("POST", url, data, oauth)

    async def put(self, url, data=None, oauth=False):
        return await self.send("PUT", url, data, oauth)

    async def delete(self, url, data=None, oauth=False):
        return await self.send("DELETE", url, data, oauth)


conn = MockConnection("test@test.com", "1234")


async def test_init(bypass_client_fetch_data):
    client = MyenergiClient(conn)
    await client.refresh()
    assert len(client.devices) == 0


async def test_init_error(error_on_client_fetch_data):
    client = MyenergiClient(conn)
    with pytest.raises(Exception):
        assert await client.refresh()


async def test_get_all_devices(client_fetch_data_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices()
    assert len(devices) == 6


async def test_get_eddi_devices(client_fetch_data_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices("eddi")
    assert len(devices) == 1
    assert isinstance(devices[0], Eddi)


async def test_get_zappi_devices(client_fetch_data_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices("zappi")
    assert len(devices) == 2
    assert isinstance(devices[1], Zappi)


async def test_get_harvi_devices(client_fetch_data_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices("harvi")
    assert len(devices) == 2
    assert isinstance(devices[1], Harvi)


async def test_1p_harvi_eddi_solar_battery(client_1p_zappi_harvi_solar_battery_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices("harvi")
    assert len(devices) == 1
    assert isinstance(devices[0], Harvi)
    devices = await client.get_devices("zappi")
    assert len(devices) == 1
    assert isinstance(devices[0], Zappi)
    assert client.power_grid == 10000
    assert client.power_generation == 5000
    assert client.power_battery == 3000
    assert client.power_charging == 2000
    assert client.consumption_home == 16000


async def test_get_libbi_devices(client_fetch_data_fixture):
    client = MyenergiClient(conn)
    devices = await client.get_devices("libbi")
    assert len(devices) == 1
    assert isinstance(devices[0], Libbi)



from pymyenergi.connection import Connection

from . import LIBBI
from .base_device import BaseDevice

STATES = {
    0: "Off",
    1: "On",
    2: "Battery Full",
    4: "Idle",
    5: "Charging",
    6: "Discharging",
    7: "Duration Charging",
    8: "Duration Drain",
    12: "Target Charge",
    51: "Boosting",
    53: "Boosting",
    55: "Boosting",
    11: "Stopped",
    101: "Battery Empty",
    102: "Full",
    104: "Full",
    151: "FW Upgrade (ARM)",
    156: "FW Upgrade (DSP)",
    172: "BMS Charge Temperature Low",
    176: "BMS Updating",
    234: "Calibration Charge",
    251: "FW Upgrade (DSP)",
    252: "FW Upgrade (ARM)",
    253: "BMS Upgrading",
}

LIBBI_MODES = ["Stopped", "Normal", "Export"]
LIBBI_MODE_CONFIG = {
    "Stopped": {"mode_int": 0, "mode_name": "STOP"},
    "Normal": {"mode_int": 1, "mode_name": "BALANCE"},
    "Export": {"mode_int": 5, "mode_name": "DRAIN"},
}
"""The myenergi app defines other modes as well (capture, charge, match), but these cannot be set"""


class Libbi(BaseDevice):
    """Libbi Client for myenergi API."""

    def __init__(self, connection: Connection, serialno, data={}) -> None:
        self.history_data = {}
        self._extra_data = {}
        super().__init__(connection, serialno, data)

    async def refresh_extra(self):
        # only refresh this data if we have app credentials
        if self._connection.app_email and self._connection.app_password:
            chargeFromGrid = await self._connection.get(
                "/api/AccountAccess/LibbiMode?serialNo=" + str(self.serial_number),
                oauth=True,
            )
            self._extra_data["charge_from_grid"] = chargeFromGrid["content"][
                str(self.serial_number)
            ]
            chargeTarget = await self._connection.get(
                "/api/AccountAccess/" + str(self.serial_number) + "/LibbiChargeSetup",
                oauth=True,
            )
            self._extra_data["charge_target"] = chargeTarget["content"]["energyTarget"]

    @property
    def kind(self):
        return LIBBI

    @property
    def status(self):
        """Get current known status"""
        n = self._data.get("sta", 1)
        if n in STATES:
            return STATES[n]
        else:
            return n

    @property
    def local_mode(self):
        """Get current known status"""
        return self._data.get("lmo", 1)

    @property
    def ct_keys(self):
        """Return CT key names that are not none"""
        keys = {}
        for i in range(6):
            ct = getattr(self, f"ct{i+1}")
            if ct.name_as_key == "ct_none":
                continue
            keys[ct.name_as_key] = keys.get(ct.name_as_key, 0) + 1
        return keys

    @property
    def ct3(self):
        """Current transformer 3"""
        return self._create_ct(3)

    @property
    def ct4(self):
        """Current transformer 4"""
        return self._create_ct(4)

    @property
    def ct5(self):
        """Current transformer 4"""
        return self._create_ct(5)

    @property
    def ct6(self):
        """Current transformer 4"""
        return self._create_ct(6)

    @property
    def supply_frequency(self):
        """Supply frequency in Hz"""
        return self._data.get("frq")

    @property
    def supply_voltage(self):
        """Supply voltage in V"""
        return self._data.get("vol", 0) / 10

    @property
    def consumed_session(self):
        """Energy diverted this session kWh"""
        return self._data.get("che", 0)

    @property
    def power_grid(self):
        """Grid power in W"""
        return self._data.get("grd", 0)

    @property
    def power_generated(self):
        """Generated power in W"""
        return self._data.get("gen", 0)

    @property
    def energy_total(self):
        """Device total energy from history data"""
        return self.history_data.get("device_total", 0)

    @property
    def energy_green(self):
        """Device green energy from history data"""
        return self.history_data.get("device_green", 0)

    @property
    def state_of_charge(self):
        """State of Charge in %"""
        return self._data.get("soc", 0)

    @property
    def priority(self):
        """Current priority"""
        return self._data.get("pri")

    @property
    def battery_size(self):
        """Battery size in kwh"""
        return self._data.get("mbc", 0) / 1000

    @property
    def inverter_size(self):
        """Inverter size in kwh"""
        return self._data.get("mic", 0) / 1000

    @property
    def grid_import(self):
        """Grid import from history data"""
        return self.history_data.get("grid_import", 0)

    @property
    def grid_export(self):
        """Grid export from history data"""
        return self.history_data.get("grid_export", 0)

    @property
    def battery_charge(self):
        """Battery charge from history data"""
        return self.history_data.get("battery_charge", 0)

    @property
    def battery_discharge(self):
        """Battery discharge from history data"""
        return self.history_data.get("battery_discharge", 0)

    @property
    def generated(self):
        """Solar generation from history data"""
        return self.history_data.get("generated", 0)

    @property
    def charge_from_grid(self):
        """Is charging from the grid enabled?"""
        if self._connection.app_email and self._connection.app_password:
            return self._extra_data.get("charge_from_grid")
        else:
            return None

    @property
    def charge_target(self):
        """Libbi charge target"""
        if self._connection.app_email and self._connection.app_password:
            return self._extra_data.get("charge_target", 0) / 1000
        else:
            return None

    @property
    def prefix(self):
        return "L"

    def get_mode_description(self, mode: str):
        """Get the mode name as returned by myenergi API. E.g. Normal mode is BALANCE"""
        for k in LIBBI_MODE_CONFIG:
            if LIBBI_MODE_CONFIG[k]["mode_name"] == mode:
                return k
        return "???"

    async def set_operating_mode(self, mode: str):
        """Set operating mode"""
        print("current mode", self.get_mode_description(self._data["lmo"]))
        mode_int = LIBBI_MODE_CONFIG[mode.capitalize()]["mode_int"]
        await self._connection.get(
            f"/cgi-libbi-mode-{self.prefix}{self._serialno}-{mode_int}"
        )
        self._data["lmo"] = LIBBI_MODE_CONFIG[mode.capitalize()]["mode_name"]
        return True

    async def set_charge_from_grid(self, charge_from_grid: bool):
        """Set charge from grid"""
        if self._connection.app_email and self._connection.app_password:
            await self._connection.put(
                f"/api/AccountAccess/LibbiMode?chargeFromGrid={charge_from_grid}&serialNo={self._serialno}",
                oauth=True,
            )
            self._extra_data["charge_from_grid"] = charge_from_grid
            return True
        else:
            return False

    async def set_priority(self, priority):
        """Set device priority"""
        await self._connection.get(
            f"/cgi-set-priority-{self.prefix}{self._serialno}-{int(priority)}"
        )
        self._data["pri"] = int(priority)
        return True

    async def set_charge_target(self, charge_target: float):
        """Set charge target"""
        if self._connection.app_email and self._connection.app_password:
            await self._connection.put(
                f"/api/AccountAccess/{self._serialno}/TargetEnergy?targetEnergy={charge_target}",
                oauth=True,
            )
            self._extra_data["charge_target"] = charge_target
            return True
        else:
            return False

    def show(self, short_format=False):
        """Returns a string with all data in human readable format"""
        ret = ""
        name = ""
        if self.name:
            name = f" {self.name}"
        ret = ret + f"Libbi S/N {self.serial_number}"
        ret = ret + f"{name}"
        if short_format:
            return ret
        ret = ret.center(80, "-") + "\n"
        ret = ret + f"Libbi priority: {self.priority}\n"
        ret = ret + f"Battery size: {self.battery_size}kWh\n"
        ret = ret + f"Inverter size: {self.inverter_size}kWh\n"
        ret = ret + f"State of Charge: {self.state_of_charge}%\n"
        ret = ret + f"Generating: {self.power_generated}W\n"
        ret = ret + f"Grid: {self.power_grid}W\n"
        ret = ret + f"Status: {self.status}\n"
        ret = ret + "Local Mode: " + self.get_mode_description(self.local_mode) + "\n"
        ret = ret + "Charge from Grid: "
        if self.charge_from_grid is not None:
            if self.charge_from_grid:
                ret = ret + "Enabled\n"
            else:
                ret = ret + "Disabled\n"
        else:
            ret = ret + "<unavailable>\n"
        ret = ret + "Charge target: "
        if self.charge_target is not None:
            ret = ret + f"{self.charge_target}kWh\n"
        else:
            ret = ret + "<unavailable>\n"
        ret = ret + f"CT 1 {self.ct1.name} {self.ct1.power}W phase {self.ct1.phase}\n"
        ret = ret + f"CT 2 {self.ct2.name} {self.ct2.power}W phase {self.ct2.phase}\n"
        ret = ret + f"CT 3 {self.ct3.name} {self.ct3.power}W phase {self.ct3.phase}\n"
        ret = ret + f"CT 4 {self.ct4.name} {self.ct4.power}W phase {self.ct4.phase}\n"
        ret = ret + f"CT 5 {self.ct5.name} {self.ct5.power}W phase {self.ct5.phase}\n"
        ret = ret + f"CT 6 {self.ct6.name} {self.ct6.power}W phase {self.ct6.phase}\n"
        for key in self.ct_keys:
            ret = ret + f"Energy {key} {self.history_data.get(key, 0)}Wh\n"
        if not self._connection.app_email or not self._connection.app_password:
            ret += "No app credentials provided - the above information might not be totally accurate\n"
        return ret
